from abc import ABCMeta, abstractmethod

from web3 import HTTPProvider, Web3
from web3.contract import Contract
from web3.eth import Eth

from atlantiscore.types import EVMAddress, PrivateKey

ETHEREUM_MAINNET_CHAIN_ID = 1
SEPOLIA_CHAIN_ID = 11155111


class EVMMiddleware:
    web3: Web3
    private_key: PrivateKey
    chain_id: int
    compatible_chain_ids: tuple[int] = (ETHEREUM_MAINNET_CHAIN_ID, SEPOLIA_CHAIN_ID)

    def __init__(
        self,
        http_node_provider_url: str,
        chain_id: int,
        private_key: PrivateKey,
    ) -> None:
        if chain_id not in self.compatible_chain_ids:
            raise ValueError("Unsupported chain_id")

        self.chain_id = chain_id
        self.web3 = Web3(HTTPProvider(http_node_provider_url))
        self.private_key = private_key

    @property
    def eth(self) -> Eth:
        return self.web3.eth

    async def create_transaction_data(self, additional_data: dict = {}) -> dict:
        nonce = await self.get_nonce()
        return {
            "chainId": self.chain_id,
            "nonce": nonce,
            "from": str(self.private_key.public_address),
            **additional_data,
        }

    async def send_unsigned_transaction(self, transaction: dict) -> str:
        signed_transaction = self.eth.account.sign_transaction(
            transaction,
            private_key=self.private_key.to_hex(),
        )
        return self.eth.send_raw_transaction(signed_transaction.rawTransaction).hex()

    async def get_nonce(self) -> int:
        return self.eth.get_transaction_count(str(self.private_key.public_address))


class EVMContractMiddleware(EVMMiddleware, metaclass=ABCMeta):
    address: EVMAddress

    def __init__(self, address: EVMAddress, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.address = address

    @staticmethod
    @abstractmethod
    async def get_abi() -> dict:
        """Returns abi for associated contract."""

    async def get_contract(self) -> Contract:
        abi = await self.get_abi()
        return self.eth.contract(address=str(self.address), abi=abi)



from atlantiscore.lib.abi import read_abi
from atlantiscore.middleware.blockchain.evm import EVMContractMiddleware
from atlantiscore.types import EVMAddress


class ERC20Middleware(EVMContractMiddleware):
    address: EVMAddress

    @staticmethod
    async def get_abi() -> dict:
        return await read_abi("erc20")

    async def transfer(
        self,
        recipient: EVMAddress,
        amount: int,
        additional_transaction_data: dict = {},
    ) -> str:
        contract = await self.get_contract()
        transfer = contract.functions.transfer(str(recipient), amount)
        tx = transfer.build_transaction(
            await self.create_transaction_data(additional_transaction_data),
        )
        return (await self.send_unsigned_transaction(tx)).hex()

    async def get_decimals(self):
        contract = await self.get_contract()
        return contract.functions.decimals().call()



import json
from pathlib import Path

import aiofiles


async def read_abi(name: str) -> dict:
    path = Path(__file__).parent / f"{name}.abi"
    async with aiofiles.open(path, mode="r") as file:
        return json.loads(await file.read())



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, Self, Type

from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic_core import core_schema

from atlantiscore.lib.exceptions import InvalidByteEncoding

BYTE_ORDER = "big"
NUMBER_OF_BITS_IN_BYTE = 8
PREFIX = "0x"
PREFIX_SIZE = len(PREFIX)

LiteralByteEncoding = str | int | bytes


class ByteEncoding(metaclass=ABCMeta):
    _example: str
    _byte_count: int
    _max_str_length: int
    _min_str_length: int
    _value: bytes

    def __bytes__(self) -> bytes:
        return self._value

    def __int__(self) -> int:
        return int.from_bytes(self._value, BYTE_ORDER)

    def __str__(self) -> str:
        return PREFIX + self._value.hex()

    @abstractmethod
    def __eq__(self, other: any) -> bool:
        """Compares this object with other and returns whether they're equal."""

    def __ne__(self, other: any) -> bool:
        return not self.__eq__(other)

    @abstractmethod
    def __hash__(self) -> int:
        """Returns a hash of the byte value."""

    def __bool__(self) -> bool:
        return bool(int(self))

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Type[Any],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(
            cls._validate,
            schema=core_schema.union_schema(
                (
                    core_schema.str_schema(
                        min_length=cls._min_str_length,
                        max_length=cls._max_str_length,
                    ),
                    core_schema.int_schema(ge=0),
                    core_schema.bytes_schema(
                        min_length=cls._byte_count,
                        max_length=cls._byte_count,
                    ),
                    core_schema.is_instance_schema(cls=cls),
                )
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: str(v),
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )

    @classmethod
    def _validate(cls, v: ByteEncoding | LiteralByteEncoding) -> Self:
        try:
            return cls(v)
        except InvalidByteEncoding as e:
            raise ValueError(str(e))

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: core_schema.CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict:
        json_schema = handler.resolve_ref_schema(handler(core_schema))
        json_schema.update(example=cls._example)
        return json_schema


def _encoding_to_bytes(
    value: ByteEncoding | LiteralByteEncoding,
    byte_count: int,
) -> bytes:
    try:
        if isinstance(value, ByteEncoding):
            return bytes(value)

        if isinstance(value, str):
            byte_sequence = _hex_to_bytes(value)
        elif isinstance(value, int):
            byte_sequence = _int_to_bytes(value, byte_count)
        elif isinstance(value, bytes):
            byte_sequence = value
        else:
            raise TypeError

        if len(byte_sequence) != byte_count:
            raise ValueError

        return byte_sequence
    except (ValueError, TypeError) as e:
        raise InvalidByteEncoding(value) from e


def _hex_to_bytes(hex_str: str) -> bytes:
    if hex_str[:PREFIX_SIZE].lower() == PREFIX:
        hex_str = hex_str[PREFIX_SIZE:]
    return bytes.fromhex(hex_str)


def _int_to_bytes(integer: int, byte_count: int) -> bytes:
    return integer.to_bytes(_calculate_required_byte_count(integer), BYTE_ORDER).rjust(
        byte_count, b"\x00"
    )


def _calculate_required_byte_count(integer: int) -> int:
    """Returns the minimum number of bytes required to represent the given int.

    Example:
    0 + 7 would require 0 bytes
    1 + 7 would require 1 byte
    """
    return (integer.bit_length() + 7) // NUMBER_OF_BITS_IN_BYTE



from __future__ import annotations

from atlantiscore.lib.exceptions import InvalidByteEncoding, InvalidEVMTransactionHash
from atlantiscore.types.evm.base import (
    PREFIX_SIZE,
    ByteEncoding,
    LiteralByteEncoding,
    _encoding_to_bytes,
)

TRANSACTION_HASH_BYTE_COUNT = 32
EXAMPLE_TRANSACTION_HASH_STRING = (
    "0xd4fe60962208702e0e5915e0268847709610d7b3b8be39b2af57ccca6809951a"
)


class EVMTransactionHash(ByteEncoding):
    _example: str = EXAMPLE_TRANSACTION_HASH_STRING
    _byte_count: int = TRANSACTION_HASH_BYTE_COUNT
    _min_str_length: int = _byte_count * 2
    _max_str_length: int = _min_str_length + PREFIX_SIZE

    def __init__(self, value: EVMTransactionHash | LiteralByteEncoding) -> None:
        self._value = _transaction_hash_to_bytes(value)

    def __eq__(self, other: any) -> bool:
        try:
            return hash(self) == hash(EVMTransactionHash(other))
        except InvalidByteEncoding:
            return False

    def __hash__(self) -> int:
        return int(self)


def _transaction_hash_to_bytes(
    value: EVMTransactionHash | LiteralByteEncoding,
) -> bytes:
    try:
        return _encoding_to_bytes(value, TRANSACTION_HASH_BYTE_COUNT)
    except InvalidByteEncoding as e:
        raise InvalidEVMTransactionHash(value) from e



from sqlalchemy.types import LargeBinary

from atlantiscore.db.types.evm.base import ByteEncoding
from atlantiscore.types.evm import EVMAddress as PythonEVMAddress, LiteralByteEncoding

BYTE_COUNT = 20


class EVMAddress(ByteEncoding):
    _default_type: LargeBinary = LargeBinary(BYTE_COUNT)

    @staticmethod
    def _parse(value: PythonEVMAddress | LiteralByteEncoding) -> PythonEVMAddress:
        return PythonEVMAddress(value)



import sys
from importlib import import_module
from importlib.abc import MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import find_spec
from types import ModuleType
from typing import Any, Optional


class AtlantisFinder(MetaPathFinder):
    finding_in_atlantis: bool = False

    @classmethod
    def find_spec(cls, fullname: str, *args, **kwargs) -> Optional[ModuleSpec]:
        if "atlantiscore" not in fullname:
            return None

        if cls.finding_in_atlantis:
            return None

        cls.finding_in_atlantis = True
        if find_spec(fullname):
            cls.finding_in_atlantis = False
            return None

        cls.finding_in_atlantis = False
        return ModuleSpec(fullname, cls)

    @classmethod
    def create_module(cls, spec: ModuleSpec) -> Any:
        redirected_name = spec.name.replace(
            "atlantiscore",
            "codercore",
        )
        try:
            return import_module(redirected_name)
        except ModuleNotFoundError as e:
            try:
                return cls.import_attribute(*redirected_name.rsplit(".", 1))
            except (AttributeError, ModuleNotFoundError, TypeError):
                raise e

    @staticmethod
    def import_attribute(module_name: str, attr_name: str) -> Any:
        return getattr(import_module(module_name), attr_name)

    @staticmethod
    def exec_module(module: ModuleType) -> None:
        # Intentionally left blank.
        pass


sys.meta_path.append(AtlantisFinder())
